#!/usr/bin/env python3
"""
Copyright 2022 Byron Labs S.L.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# TODO Referenciar vt-py

from datetime import datetime, timedelta
import logging
from typing import Union
from urllib.parse import urljoin, urlencode

# from pydantic import validate_arguments
import requests
from vysion.client.error import APIError

import vysion.model as model
from vysion.model import VysionError
from vysion.version import __version__ as vysion_version

_API_HOST = "https://api.vysion.ai"

# All API endpoints start with this prefix, you don't need to include the
# prefix in the paths you request as it's prepended automatically.
_ENDPOINT_PREFIX = "/api/v1/"

_BASE_API = urljoin(_API_HOST, _ENDPOINT_PREFIX)

LOGGER = logging.getLogger("vysion-py")
LOGGER.setLevel(logging.INFO)

# __all__ = []


class BaseClient:

    # __attrs__ = []

    # @validate_arguments #
    def __init__(self, api_key: str = None, headers: dict = dict(), proxy: dict = None):

        assert api_key is not None, "API key MUST be provided"
        assert isinstance(api_key, str), "API key MUST be a string"

        self.api_key = api_key
        self.proxy = proxy  # TODO Rename proxy to proxies
        self.headers = headers

    def __get_session__(self) -> requests.Session:

        # TODO Configure proxy

        # If session is undefined
        try:
            self._session
        except (NameError, AttributeError):

            headers = self.headers.copy()
            headers.update(
                {
                    "X-API-KEY": self.api_key,
                    "User-Agent": "vysion-py/%s" % vysion_version,
                }
            )

            self._session: requests.Session = requests.Session()
            self._session.headers.update(headers)
            self._session.proxies = self.proxy

        return self._session

    def _build_api_url__(self, endpoint, param, **query_params):

        base = urljoin(_BASE_API, f"{endpoint}/{param}")

        query_params_initialzed = query_params.copy()

        keys = list(query_params.keys())
        keys.sort()

        for i in keys:

            v = query_params[i]

            if v is None:
                del query_params_initialzed[i]

        query = "?" + urlencode(query_params_initialzed)

        return urljoin(base, query)

    def _make_request(self, url: str) -> model.VysionResponse:

        session = self.__get_session__()
        r = session.get(url)

        # TODO Improve this
        if r.status_code != 200:
            try:
                err = r.json()
                code = err.get("code")
                message = err.get("message")
            except:
                code = r.status_code
                message = r.text

            raise APIError(code, message)

        payload = r.json()

        result = model.VysionResponse.parse_obj(payload)

        return result


def vysion_error_manager(method) -> Union[model.Result, VysionError]:

    def manage(*args, **kwargs):
        try:
            result = method(*args, **kwargs)
            return result
        except APIError as e:
            return VysionError(code=e.code, message=e.message)
        except Exception as e:
            LOGGER.error(e)
            return VysionError()

    return manage


class Client(BaseClient):

    # def add_url(self, url:str, type:VysionURL.Type):
    #     """Add a Tor URL to be analyzed by PARCHE.

    #     :param url: URL to be scanned.
    #     :param type: Instance of :class:`VysionURL.Type`
    #     :returns: An instance of :class:`VysionResponse`
    #     """
    #     pass

    # def find_onion(self):
    #   pass

    # def add_onion(self):
    #   pass

    # def consume_feed(self):
    #   pass

    @vysion_error_manager
    def status(self):
        # TODO Check API status
        pass

    @vysion_error_manager
    def search(
        self,
        query: str,
        exact: bool = False,
        network: model.Network = None,
        language: model.Language = None,
        page: int = 1,
        before: datetime = None,
        after: datetime = None,
    ) -> model.Result:

        url = self._build_api_url__(
            "search",
            query,
            exact=exact,
            network=network,
            language=language,
            page=page,
            before=before,
            after=after,
        )

        result = self._make_request(url)
        return result.data

    @vysion_error_manager
    def find_btc(
        self, btc: str, page: int = 1, before: datetime = None, after: datetime = None
    ) -> model.Result:

        url = self._build_api_url__("btc", btc, page=page, before=before, after=after)

        result = self._make_request(url)
        return result.data

    # TODO find_domain?
    @vysion_error_manager
    def find_url(
        self,
        query_url: str,
        page: int = 1,
        before: datetime = None,
        after: datetime = None,
    ) -> model.Result:

        url = self._build_api_url__(
            "url", query_url, page=page, before=before, after=after
        )

        result = self._make_request(url)
        return result.data

    @vysion_error_manager
    def find_email(
        self, email: str, page: int = 1, before: datetime = None, after: datetime = None
    ) -> model.Result:

        url = self._build_api_url__(
            "email", email, page=page, before=before, after=after
        )

        result = self._make_request(url)
        return result.data

    @vysion_error_manager
    def get_document(self, document_id: str) -> model.Result:

        url = self._build_api_url__("document", document_id)

        result = self._make_request(url)
        return result.data

    # FEEDS
    @vysion_error_manager
    def consume_feed_ransomware(self, batch_day: datetime = datetime.today()):
        pass


# TODO Develop feeds logic
# Example: https://github.com/VirusTotal/vt-py/blob/master/vt/feed.py
class DaylyFeed(Client):
    def _consume_batch(self, start_time, end_time):
        raise NotImplementedError()

    def consume(self, batch_day: datetime = datetime.today()):
        start_time = datetime(datetime.year, datetime.month, datetime.day)
        end_time = start_time + timedelta(days=1)
        return self._consume_batch(start_time, end_time)


class RansomwareFeed(DaylyFeed):
    def _consume_batch(self, start_time, end_time):

        days = (datetime.now() - start_time).days
        pages = (end_time - start_time).days

        for page in range(pages):
            url = self._build_api_url__("feed", "ransomware", days=days, page=page + 1)
            yield self._make_request(url)


"""TODO Transform response
{
  "total": {
    "value": 26,
    "relation": "eq"
  },
  "max_score": null,
  "hits": [
    {
      "_index": "ransomware-62022",
      "_id": "dab4447fbd8440251d41e31b8ab1770186f429e46d9cd6903fbe17367efcba23",
      "_score": null,
      "_ignored": [
        "info.keyword"
      ],
      "_source": {
        "company": "RG Alliance Group",
        "link": "http://quantum445bh3gzuyilxdzs5xdepf3b7lkcupswvkryf3n7hgzpxebid.onion/target/rgalliancegroup",
        "date": "2022-06-16T00:00:00",
        "group": "Quantum",
        "info": "1.2TB 3.6K visibility\n2022-06-16\nInfo Post About: \"RG Alliance Group\"\nCompany Name\nRG Alliance Group\nCompany Website\nOfficial Link\nTotal Revenue\n$9M\nLast Updated:\n2022-06-16\nVolume Of Data Uploaded\n0% (announcement)\n- Complete databases of clients of the accounting program QuickBooks, including tax, payroll and banking information of Client Companies.\n- Contracts, Accounting.Ballances, Scans, invoices, etc. Client companies\n- Email correspondence with client companies\n- Personal information (Scans+SSN+Bank info+etc) of employees of Client Companies\n- Personal information (Scans+SSN+Bank info+etc) of RG Alliance Group employees\n- Outlook .pst backups of RG Alliance Group employees\nYour Feedback\nAuthor\nMessage\nSUBMIT",
        "company_link": "http://www.rgalliance.com/"
      },
      "sort": [
        1655337600000
      ]
    }
    ...
  ]
}
"""

# TODO /api/v1/feeds



#!/usr/bin/env python3
"""
Copyright 2022 ByronLabs S.L.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from enum import Enum
import hashlib

from datetime import datetime

try:
    from types import NoneType
except:
    NoneType: type = type(None)

from typing import List, Optional, Union
from urllib.parse import urlparse

from pydantic import BaseModel, Field # , constr

from vysion import taxonomy as vystaxonomy
from .enum import Services, Network, Language, RansomGroup


class Email(BaseModel):

    _taxonomy = [vystaxonomy.Email]

    # RFC 5322 Official Standard (https://www.emailregex.com/)
    # value: constr(regex=r'''(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])''') # TODO Añadir que es str
    value: str  # TODO Fix regex to allow caps


class Paste(BaseModel):

    _taxonomy = [vystaxonomy.Pastebin, vystaxonomy.JustPaste]

    value: str  # TODO Regex


class Skype(BaseModel):

    _taxonomy = [vystaxonomy.Skype]

    value: str  # TODO Regex


class Telegram(BaseModel):

    _taxonomy = [vystaxonomy.Telegram]  # TODO Create Telegram URL

    value: str  # TODO Regex


class BitcoinAddress(BaseModel):

    _taxonomy = [vystaxonomy.Bitcoin_Address]  # TODO Create Telegram URL

    value: str  # TODO Regex


class WhatsApp(BaseModel):

    _taxonomy = [vystaxonomy.WhatsApp]

    value: str  # TODO Regex


class URL(BaseModel):

    _taxonomy = [vystaxonomy.URL]

    protocol: str
    domain: str
    port: int
    path: str
    signature: str
    network: Network

    @classmethod
    def parse(cls, url):

        parsed = urlparse(url)

        scheme = parsed.scheme
        netloc = parsed.netloc
        path = parsed.path
        query = parsed.query
        fragment = parsed.fragment
        params = parsed.params
        username = parsed.username
        password = parsed.password

        # Build domain:port
        domain_port = netloc.split(":")
        domain = domain_port[0]
        if len(domain_port) <= 1:
            try:
                port = Services[scheme]
            except KeyError:
                port = 80
        else:
            port = domain_port[1]

        # Rebuild path's query
        query_parts = [param.split("=") for param in query.split("&")]
        query_dict = {}
        for part in query_parts:
            if len(part) <= 1:
                query_dict[part[0]] = str()
            else:
                query_dict[part[0]] = part[1]

        query_keys = list(query_dict.keys())
        query_keys.sort()
        res_query_parts = [f"{k}={query_dict[k]}" for k in query_keys]
        res_query = "?" + "&".join(res_query_parts)

        # Build /path?query#fragment
        res_path = path + res_query + f"#{fragment}"

        # TODO Adapt restalker.link_extractors.UUF logic to fix URLs
        # TODO Detect network
        tmp_result = cls(
            protocol=scheme,
            domain=domain,
            port=port,
            path=res_path,
            signature=str(),
            network=Network.clearnet,
        )

        signature = hashlib.sha1(tmp_result.build().encode()).hexdigest()
        tmp_result.signature = signature

        return tmp_result

    def build(self) -> str:
        return f"{self.protocol}://{self.domain}:{self.port}{self.path}"


class Page(BaseModel):

    id: str
    url: URL
    parent: str = None
    title: str = None  # TODO Revisar si None o str()
    language: Language
    html: str = None
    sha1sum: str = None
    ssdeep: str = None
    date: datetime = None
    chunk: bool = False


class Hit(BaseModel):

    page: Page
    email: List[Email] = Field(default_factory=lambda: [])
    paste: List[Paste] = Field(default_factory=lambda: [])
    skype: List[Skype] = Field(default_factory=lambda: [])
    telegram: List[Telegram] = Field(default_factory=lambda: [])
    whatsapp: List[WhatsApp] = Field(default_factory=lambda: [])
    bitcoin_address: List[BitcoinAddress] = Field(default_factory=lambda: [])


class RansomFeedHit(BaseModel):

    id: str
    company: Optional[str]
    company_link: Optional[str]
    link: str
    group: RansomGroup
    date: datetime
    info: Optional[str]


class Result(BaseModel):

    # TODO Añadir paginación, query, etc?
    total: int = 0
    hits: Union[List[Hit], List[RansomFeedHit]] = Field(default_factory=lambda: [])

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.total <= 0:
            self.total = len(self.hits)

    def get_type(self) -> type:

        if len(self.hits) <= 0:
            return NoneType

        return type(self.hits[0])


# TODO Move API responses to other class
class VysionResponse(BaseModel):
    """
    VysionResponse is a json:api flavoured response from the API
    """

    data: Result  # TODO Add type to all JSON:API responses


class VysionError(BaseModel):
    class StatusCode(int, Enum):

        UNK = 000
        OK = 200
        REQ_ERROR = 400
        UNAUTHORIZED = 403
        NOT_FOUND = 404
        INTERNAL_ERROR = 500

    code: StatusCode = StatusCode.UNK
    message: str = "UNK_ERR"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vysion', 'vysion.client', 'vysion.model', 'vysion.taxonomy']

package_data = \
{'': ['*']}

install_requires = \
['pydantic==1.9.1', 'pymisp==2.4.159', 'requests==2.27.1']

setup_kwargs = {
    'name': 'vysion',
    'version': '0.3.2',
    'description': 'The official Python client library for Vysion',
    'long_description': "\n# References\n\n- https://pydantic-docs.helpmanual.io/\n- https://python-patterns.guide/\n- https://docs.pytest.org/en/7.1.x/getting-started.html#get-started\n- https://github.com/samuelcolvin/pydantic\n- https://python-poetry.org/docs/pyproject\n- https://github.com/wemake-services/wemake-python-package\n\n## Style\n\n- https://code.visualstudio.com/docs/languages/python\n- https://code.visualstudio.com/docs/python/linting\n- https://code.visualstudio.com/api/language-extensions/snippet-guide\n- https://semver.org/\n- https://snyk.io/blog/python-poetry-package-manager/\n\n### Taxonomy\n\n- https://github.com/MISP/misp-taxonomies/blob/main/tools/machinetag.py\n- https://www.misp-project.org/taxonomies.html#_mapping_of_taxonomies\n- https://github.com/MISP/misp-taxonomies\n- https://github.com/MISP/misp-taxonomies/blob/main/tools/docs/images/taxonomy-explanation.png\n\n## Entorno\n\n- vs-code\n    - Extensiones:\n        - snyk\n        - flake8\n\n## The Zen of Python\n\n```python\nimport this\n> The Zen of Python, by Tim Peters\n> \n> Beautiful is better than ugly.\n> Explicit is better than implicit.\n> Simple is better than complex.\n> Complex is better than complicated.\n> Flat is better than nested.\n> Sparse is better than dense.\n> Readability counts.\n> Special cases aren't special enough to break the rules.\n> Although practicality beats purity.\n> Errors should never pass silently.\n> Unless explicitly silenced.\n> In the face of ambiguity, refuse the temptation to guess.\n> There should be one-- and preferably only one --obvious way to do it.\n> Although that way may not be obvious at first unless you're Dutch.\n> Now is better than never.\n> Although never is often better than *right* now.\n> If the implementation is hard to explain, it's a bad idea.\n> If the implementation is easy to explain, it may be a good idea.\n> Namespaces are one honking great idea -- let's do more of those!\n```\n\n# Ackowledgement\n\nWe really appreciate the documentation of https://github.com/VirusTotal/vt-py\n",
    'author': 'Javier Junquera-Sánchez',
    'author_email': 'javier.junquera@byronlabs.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://vysion.ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)
